import datetime
import os
import unittest
import inspect

import cf

class RegridTest(unittest.TestCase):
    filename1 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file1.nc')
    filename2 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file2.nc')
    filename3 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file3.nc')
    filename4 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file4.nc')
    filename5 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'test_file3.nc')
    filename6 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'test_file2.nc')
    filename7 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file5.nc')
    filename8 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file6.nc')
    filename9 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file7.nc')
    filename10 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file8.nc')

    chunk_sizes = (300, 10000, 100000)[::-1]
    original_chunksize = cf.CHUNKSIZE()

    test_only = []
#    test_only = ('NOTHING!!!!!',)
#    test_only = ('test_Field_regrids',)
#    test_only = ('test_Field_regridc',)
#    test_only('test_Field_section',)
#    test_only('test_Data_section',)

    @unittest.skipUnless(cf._found_ESMF, "Requires esmf package.")
    def test_Field_regrids(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Set tolerance for array equality
        original_atol = cf.ATOL(1e-12)

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)
            f1 = cf.read(self.filename1)[0]
            f2 = cf.read(self.filename2)[0]
            f3 = cf.read(self.filename3)[0]

            r = f1.regrids(f2, 'conservative')

            self.assertTrue(f3.equals(r, verbose=True),
                            'destination = global Field, CHUNKSIZE = %s' % chunksize)
            r = f1.regrids(f2, method='conservative')

            self.assertTrue(f3.equals(r, verbose=True),
                            'destination = global Field, CHUNKSIZE = %s' % chunksize)
            dst = {'longitude': f2.dim('X'), 'latitude': f2.dim('Y')}
            r = f1.regrids(dst, 'conservative', dst_cyclic=True)

            self.assertTrue(f3.equals(r, verbose=True),
                            'destination = global dict, CHUNKSIZE = %s' % chunksize)
            r = f1.regrids(dst, method='conservative', dst_cyclic=True)
            self.assertTrue(f3.equals(r, verbose=True),
                            'destination = global dict, CHUNKSIZE = %s' % chunksize)

            f4 = cf.read(self.filename4)[0]
            f5 = cf.read(self.filename5)[0]

            r = f1.regrids(f5, 'linear')
            self.assertTrue(f4.equals(r, verbose=True),
                            'destination = regional Field, CHUNKSIZE = %s' % chunksize)

            r = f1.regrids(f5, method='linear')
            self.assertTrue(f4.equals(r, verbose=True),
                            'destination = regional Field, CHUNKSIZE = %s' % chunksize)
        # --- End: for
        cf.CHUNKSIZE(self.original_chunksize)

        f6 = cf.read(self.filename6)[0]
        with self.assertRaises(Exception):
            f1.regridc(f6, axes='T', method='linear')

        cf.ATOL(original_atol)


    @unittest.skipUnless(cf._found_ESMF, "Requires esmf package.")
    def test_Field_regridc(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        original_atol = cf.ATOL(1e-12)

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)
            f1 = cf.read(self.filename7)[0]
            f2 = cf.read(self.filename8)[0]
            f3 = cf.read(self.filename9)[0]
            self.assertTrue(f3.equals(f1.regridc(f2, axes='T', method='linear'), verbose=True),
                            'destination = time series, CHUNKSIZE = %s' % chunksize)
            f4 = cf.read(self.filename1)[0]
            f5 = cf.read(self.filename2)[0]
            f6 = cf.read(self.filename10)[0]
            self.assertTrue(f6.equals(f4.regridc(f5, axes=('X', 'Y'), method='conservative'),
                                      verbose=True),
                            'destination = global Field, CHUNKSIZE = %s' % chunksize)
            self.assertTrue(f6.equals(f4.regridc(f5, axes=('X', 'Y'), method='conservative'),
                                      verbose=True),
                            'destination = global Field, CHUNKSIZE = %s' % chunksize)
            dst = {'X': f5.dim('X'), 'Y': f5.dim('Y')}
            self.assertTrue(f6.equals(f4.regridc(dst, axes=('X', 'Y'), method='conservative'),
                                      verbose=True),
                            'destination = global dict, CHUNKSIZE = %s' % chunksize)
            self.assertTrue(f6.equals(f4.regridc(dst, axes=('X', 'Y'), method='conservative'),
                                      verbose=True),
                            'destination = global dict, CHUNKSIZE = %s' % chunksize)
        # --- End: for
        cf.CHUNKSIZE(self.original_chunksize)

        cf.ATOL(original_atol)


# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print('')
    unittest.main(verbosity=2)




from os import close
from operator import mul

import numpy

from ..functions import parse_indices, get_subspace

from . import abstract

_debug = False


class RaggedIndexedSubarray(abstract.CompressedSubarray):
    '''TODO

    '''
    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

    Returns a numpy array.

        '''
        # The compressed array
        array = self.array

        # Initialize the full, uncompressed output array with missing
        # data everywhere
        uarray = numpy.ma.masked_all(self.shape, dtype=array.dtype)

        r_indices = [slice(None)] * array.ndim
        p_indices = [slice(None)] * uarray.ndim

        compression = self.compression

        instance_axis = compression['instance_axis']
        instance_index = compression['instance_index']
        element_axis = compression['i_element_axis']
        sample_indices = compression['i_element_indices']

        p_indices[instance_axis] = instance_index
        p_indices[element_axis] = slice(0, len(sample_indices))

        uarray[tuple(p_indices)] = array[sample_indices]

        if _debug:
            print('instance_axis    =', instance_axis)
            print('instance_index   =', instance_index)
            print('element_axis     =', element_axis)
            print('sample_indices   =', sample_indices)
            print('p_indices        =', p_indices)
            print('uarray.shape     =', uarray.shape)
            print('self.array.shape =', array.shape)

        if indices is Ellipsis:
            return uarray
        else:
            if _debug:
                print('indices =', indices)

            indices = parse_indices(self.shape, indices)
            if _debug:
                print('parse_indices(self.shape, indices) =', indices)

            return get_subspace(uarray, indices)


#    def __repr__(self):
#        '''x.__repr__() <==> repr(x)
#
#        '''
#        return "<CF %s: %s>" % (self.__class__.__name__, str(self.array))


# --- End: class



import os
import numpy
import numpy.ctypeslib
import ctypes as CT

from . import umfile


_len_real_hdr = 19
_len_int_hdr = 45


class File_type(CT.Structure):
    _fields_ = [
        ("format", CT.c_int),
        ("byte_ordering", CT.c_int),
        ("word_size", CT.c_int)
    ]


def _get_ctypes_array(dtype, size=None):
    """
    get ctypes corresponding to a numpy array of a given type;
    the size should not be necessary unless the storage for the array
    is allocated in the C code
    """
    kwargs = {
        'dtype': dtype,
        'ndim': 1,
        'flags': ('C_CONTIGUOUS', 'WRITEABLE')
    }
    if size:
        kwargs['shape'] = (size,)
    return numpy.ctypeslib.ndpointer(**kwargs)


def _gen_rec_class(int_type, float_type):
    class Rec(CT.Structure):
        """
        ctypes object corresponding to the Rec object in the C code,
        """
        _fields_ = [
            ("int_hdr", _get_ctypes_array(int_type, _len_int_hdr)),
            ("real_hdr", _get_ctypes_array(float_type, _len_real_hdr)),
            ("header_offset", CT.c_size_t),
            ("data_offset", CT.c_size_t),
            ("disk_length", CT.c_size_t),
            ("_internp", CT.c_void_p)
        ]
    return Rec


Rec32 = _gen_rec_class(numpy.int32, numpy.float32)
Rec64 = _gen_rec_class(numpy.int64, numpy.float64)


def _gen_var_class(rec_class):
    class Var(CT.Structure):
        """
        ctypes object corresponding to the Var object in the C code
        """
        _fields_ = [
            ("recs", CT.POINTER(CT.POINTER(rec_class))),
            ("nz", CT.c_int),
            ("nt", CT.c_int),
            ("supervar_index", CT.c_int),
            ("_internp", CT.c_void_p)
        ]
    return Var


Var32 = _gen_var_class(Rec32)
Var64 = _gen_var_class(Rec64)


def _gen_file_class(var_class):
    class File(CT.Structure):
        """
        ctypes object corresponding to the File object in the C code
        """
        _fields_ = [
            ("fd", CT.c_int),
            ("file_type", File_type),
            ("nvars", CT.c_int),
            ("vars", CT.POINTER(CT.POINTER(var_class))),
            ("_internp", CT.c_void_p)
        ]
    return File


File32 = _gen_file_class(Var32)
File64 = _gen_file_class(Var64)


class Enum:
    def __init__(self, *names):
        self.names = names

    def as_name(self, val):
        if isinstance(val, str):
            return val
        else:
            return self.names[val]

    def as_index(self, val):
        if isinstance(val, int):
            return val
        return self.names.index(val)


enum_file_format = Enum("PP", "FF")
enum_byte_ordering = Enum("little_endian", "big_endian")
enum_data_type = Enum("integer", "real")


class CInterface:
    """
    Interface to the C shared library functions
    """
    def __init__(self, lib_name="umfile.so"):
        lib_dir = os.path.join(
            os.path.dirname(__file__) or ".", "c-lib")
        lib_path = os.path.join(lib_dir, lib_name)
        self.lib = CT.CDLL(lib_path)

    def _is_null_pointer(self, ptr):
        try:
            ptr.contents
            return False
        except ValueError:
            return True

    def detect_file_type(self, fd):
        """
        auto-detect file type; returns a File_type ctypes object
        that can be passed to file_parse(), or
        raises an exception if file type cannot be detected
        """
        file_type = File_type()
        rv = self.lib.detect_file_type(fd, CT.pointer(file_type))
        if rv != 0:
            raise umfile.UMFileException("File type could not be detected")
        return file_type

    def file_type_obj_to_dict(self, file_type):
        """
        converts a FileType object returned by detect_file_type()
        into a dictionary that include meaningful string
        values in place of the integers that derive from the C enum
        statments, specifically:
           'format': 'PP' or 'FF'
           'byte_ordering': 'little_endian' or 'big_endian'
        and also
           'word_size': 4 or 8
        """
        format = enum_file_format.as_name(file_type.format)
        byte_ordering = enum_byte_ordering.as_name(file_type.byte_ordering)
        word_size = file_type.word_size
        return {
            'format': format,
            'byte_ordering': byte_ordering,
            'word_size': word_size
            }

    def create_file_type(self, format, byte_ordering, word_size):
        """
        takes string input values:
           'format': 'PP' or 'FF'
           'byte_ordering': 'little_endian' or 'big_endian'
           'word_size': 4 or 8
        and returns a FileType object (ctypes structure containing integer
        values) that can be passed to file_parse()
        """
        return File_type(
            format=enum_file_format.as_index(format),
            byte_ordering=enum_byte_ordering.as_index(byte_ordering),
            word_size=word_size
        )

    def set_word_size(self, val):
        """
        Sets the word size used to interpret returned pointers from subsequent
        calls, in particular the pointers to PP headers embedded in the tree of
        objects returned by file_parse() and the data array that is populated
        by read_record_data().

        the 'val' argument contains either just the word_size value to use or a
        file_type object from which it is to be extracted
        """
        if isinstance(val, File_type):
            word_size = val.word_size
        else:
            word_size = val

        if word_size == 4:
            self.file_class = File32
            self.file_data_int_type = numpy.int32
            self.file_data_real_type = numpy.float32
            self._int_ptr = CT.POINTER(CT.c_int32)
            self._real_ptr = CT.POINTER(CT.c_float)
        elif word_size == 8:
            self.file_class = File64
            self.file_data_int_type = numpy.int64
            self.file_data_real_type = numpy.float64
            self._int_ptr = CT.POINTER(CT.c_int64)
            self._real_ptr = CT.POINTER(CT.c_double)
        else:
            raise ValueError("word size must be 4 or 8 (not %s)" % word_size)

    def _get_ctypes_int_array(self, size=None):
        return _get_ctypes_array(self.file_data_int_type, size)

    def _get_ctypes_real_array(self, size=None):
        return _get_ctypes_array(self.file_data_real_type, size)

    def _get_empty_real_array(self, size):
        """
        get empty numpy real array according to word size previously
        set with set_word_size()
        """
        return numpy.empty(size, dtype=self.file_data_real_type)

    def _get_empty_int_array(self, size):
        """
        as _get_empty_real_array but for int
        """
        return numpy.empty(size, dtype=self.file_data_int_type)

    def parse_file(self, fh, file_type):
        """
        Given an open file handle, work out information from the file, and
        return this in a dictionary, of which currently the only key actually
        implemented is 'vars', containing a list of variables, as that is all
        that the caller requires.

        arguments: fh - low-level file handle (integer)
                   file_type - File_type object as returned by
                               detect_file_type() or create_file_type()
        """
        func = self.lib.file_parse
        file_p_type = CT.POINTER(self.file_class)
        func.restype = file_p_type
        file_p = func(fh, file_type)
        if self._is_null_pointer(file_p):
            raise umfile.UMFileException("file parsing failed")
        file = file_p.contents
        c_vars = file.vars[:file.nvars]
        rv = {'vars': list(map(self.c_var_to_py_var, c_vars))}
        # now that we have copied all the data into python objects for the
        # caller, free any memory allocated in the C code before returning
        free_func = self.lib.file_free
        free_func._fields_ = file_p_type
        free_func(file_p)
        return rv

    def c_var_to_py_var(self, c_var_p):
        """
        create a umfile.Var object from a ctypes object corresponding to 'Var*'
        in the C code
        """
        c_var = c_var_p.contents
        nz = c_var.nz
        nt = c_var.nt
        svi = c_var.supervar_index
        if svi < 0:
            svi = None
        c_recs = c_var.recs
        recs = [self.c_rec_to_py_rec(c_recs[recid]) for recid in
                range(nz * nt)]
        return umfile.Var(recs, nz, nt, svi)

    def c_rec_to_py_rec(self, c_rec_p):
        """
        create a umfile.Rec object from a ctypes object corresponding to 'Rec*'
        in the C code
        """
        c_rec = c_rec_p.contents

        # ============================================================
        # Previous code - causing memory leaks per
        # https://github.com/numpy/numpy/issues/6511
        #
        #     # numpy.copy used here so we can go back and free the memory
        #     # allocated by C without affecting the python object
        #     int_hdr = numpy.copy(numpy.ctypeslib.as_array(c_rec.int_hdr))
        #     real_hdr = numpy.copy(numpy.ctypeslib.as_array(c_rec.real_hdr))
        # ============================================================

        #
        # -------------------------------
        # Workaround: instead cast to a pointer, obtain the values knowing
        # the length of the header arrays, and copy into an appropriate
        # numpy array
        #
        ptr = CT.cast(c_rec.int_hdr, self._int_ptr)
        int_hdr = numpy.array(
            ptr[:_len_int_hdr], dtype=self.file_data_int_type)

        ptr = CT.cast(c_rec.real_hdr, self._real_ptr)
        real_hdr = numpy.array(
            ptr[:_len_real_hdr], dtype=self.file_data_real_type)
        # ===============================================================

        header_offset = c_rec.header_offset
        data_offset = c_rec.data_offset
        disk_length = c_rec.disk_length
        return umfile.Rec(
            int_hdr, real_hdr, header_offset, data_offset, disk_length)

    def get_type_and_num_words(self, int_hdr):
        """
        from integer header, work out data type and number of words to read
        (read_record_data requires this)

        returns 2-tuple of:
           data type: 'integer' or 'real'
           number of words
        """
        word_size = int_hdr.itemsize
        self.lib.get_type_and_num_words.argtypes = [
            CT.c_int,
            self._get_ctypes_int_array(),
            CT.POINTER(CT.c_int),
            CT.POINTER(CT.c_size_t)
        ]
        data_type = CT.c_int()
        num_words = CT.c_size_t()
        rv = self.lib.get_type_and_num_words(
            word_size,
            int_hdr,
            CT.pointer(data_type),
            CT.pointer(num_words)
        )
        if rv != 0:
            raise umfile.UMFileException(
                "error determining data type and size from integer header")
        return enum_data_type.as_name(data_type.value), num_words.value

    def get_extra_data_offset_and_length(
            self, int_hdr, data_offset, disk_length):
        """
        from integer header, gets offset and length of extra data
        returns 2-tuple (offset, length), both in units of BYTES
        """
        word_size = int_hdr.itemsize
        func = self.lib.get_extra_data_offset_and_length
        func.argtypes = [
            CT.c_int,
            self._get_ctypes_int_array(),
            CT.c_size_t,
            CT.c_size_t,
            CT.POINTER(CT.c_size_t),
            CT.POINTER(CT.c_size_t)
        ]
        extra_data_offset = CT.c_size_t()
        extra_data_length = CT.c_size_t()
        rv = func(
            word_size, int_hdr, data_offset, disk_length,
            CT.pointer(extra_data_offset), CT.pointer(extra_data_length)
        )
        if rv != 0:
            raise umfile.UMFileException(
                "error determining extra data length from integer header")
        return extra_data_offset.value, extra_data_length.value

    def read_header(
        self,
        fd,
        header_offset,
        byte_ordering,
        word_size
    ):
        """
        reads header from open file, returning as 2-tuple (int_hdr, real_hdr)
        of numpy arrays
        """
        self.lib.read_header.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            self._get_ctypes_int_array(),
            self._get_ctypes_real_array()
        ]

        int_hdr = self._get_empty_int_array(_len_int_hdr)
        real_hdr = self._get_empty_real_array(_len_real_hdr)
        rv = self.lib.read_header(
            fd,
            header_offset,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            int_hdr,
            real_hdr
        )
        if rv != 0:
            raise umfile.UMFileException("error reading header data")

        return int_hdr, real_hdr

    def read_extra_data(
        self,
        fd,
        extra_data_offset,
        extra_data_length,
        byte_ordering,
        word_size
    ):
        """
        reads record data from open file

        inputs:
           fd - integer low-level file descriptor
           extra_data_offset - offset in bytes
           extra_disk_length - disk length of extra data in bytes
           byte_ordering - 'little_endian' or 'big_endian'
           word_size - 4 or 8

        returns: extra data as string)
        """

        extra_data = b"\0" * extra_data_length

        self.lib.read_extra_data.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            CT.c_char_p
        ]

        rv = self.lib.read_extra_data(
            fd,
            extra_data_offset,
            extra_data_length,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            extra_data
        )
        if rv != 0:
            raise umfile.UMFileException("error reading extra data")

        return extra_data

    def read_record_data(
        self,
        fd,
        data_offset,
        disk_length,
        byte_ordering,
        word_size,
        int_hdr,
        real_hdr,
        data_type,
        nwords
    ):
        """
        reads record data from open file

        inputs:
           fd - integer low-level file descriptor
           data_offset - offset in words
           disk_length - disk length of data record in words
           byte_ordering - 'little_endian' or 'big_endian'
           word_size - 4 or 8
           int_hdr - integer PP headers (numpy array)
           real_hdr - real PP headers (numpy array)
           data_type - 'integer' or 'real'
           nwords - number of words to read
              type and nwords should have been returned by
              get_type_and_num_words()
        """
        if data_type == 'integer':
            data = self._get_empty_int_array(nwords)
            ctypes_data = self._get_ctypes_int_array()
        elif data_type == 'real':
            data = self._get_empty_real_array(nwords)
            ctypes_data = self._get_ctypes_real_array()
        else:
            raise ValueError("data_type must be 'integer' or 'real'")

        self.lib.read_record_data.argtypes = [
            CT.c_int,
            CT.c_size_t,
            CT.c_size_t,
            CT.c_int,
            CT.c_int,
            self._get_ctypes_int_array(),
            self._get_ctypes_real_array(),
            CT.c_size_t,
            ctypes_data
        ]

        rv = self.lib.read_record_data(
            fd,
            data_offset,
            disk_length,
            enum_byte_ordering.as_index(byte_ordering),
            word_size,
            int_hdr,
            real_hdr,
            nwords,
            data
        )

        if rv != 0:
            raise umfile.UMFileException("error reading record data")

        return data


if __name__ == "__main__":
    import sys

    c = CInterface()
    fd = os.open(sys.argv[1], os.O_RDONLY)
    file_type = c.detect_file_type(fd)
    print(c.file_type_obj_to_dict(file_type))
    c.set_word_size(file_type)
    info = c.parse_file(fd, file_type)

    for var in info['vars']:
        print("nz = %s, nt = %s" % (var.nz, var.nt))
        for rec in var.recs:
            print(rec.hdr_offset)
            print("data offset", rec.data_offset)
            print("disk length", rec.disk_length)
            print("int hdr", rec.int_hdr)
            print("real hdr", rec.real_hdr)
            data_type, nwords = c.get_type_and_num_words(rec.int_hdr)
            print("data_type = %s nwords = %s" % (data_type, nwords))

            data = c.read_record_data(
                fd,
                rec.data_offset,
                rec.disk_length,
                file_type.byte_ordering,
                file_type.word_size,
                rec.int_hdr,
                rec.real_hdr,
                data_type,
                nwords
            )
            print("data (%s values): %s ... %s" % (
                len(data), data[:10], data[-10:]))
            extra_data_offset, extra_data_length = (
                c.get_extra_data_offset_and_length(
                    rec.int_hdr,
                    rec.data_offset,
                    rec.disk_length
                )
            )
            print("extra data offset: %s" % extra_data_offset)
            print("extra data length: %s" % extra_data_length)
            extra_data = c.read_extra_data(
                fd,
                extra_data_offset,
                extra_data_length,
                file_type.byte_ordering,
                file_type.word_size
            )
            print("extra data (%s bytes) read" % (len(extra_data)))

    print(c.read_header(
        fd,
        info['vars'][0].recs[0].hdr_offset,
        file_type.byte_ordering,
        file_type.word_size
    ))



import cfdm

from .cfimplementation import implementation


_implementation = implementation()

def example_field(n, _implementation=_implementation):
    return cfdm.example_field(n, _implementation=_implementation)

example_field.__doc__ = cfdm.example_field.__doc__



import cfdm

from . import mixin


class NodeCountProperties(mixin.Properties,
                          cfdm.NodeCountProperties):
    '''Properties for a netCDF node count variable.

    **NetCDF interface**

    The netCDF node count variable name may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
    `nc_has_variable` methods.

    .. versionadded:: 3.2.0

    '''
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

# --- End: class



import copy
import datetime
import re
import os
import unittest

import numpy

import cf

class QueryTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    def test_Query_contains(self):
        f = cf.read(self.filename)[0]
        c = f.dim('X')
        self.assertTrue(((cf.contains(21.1)==c).array ==
                         numpy.array([0,1,0,0,0,0,0,0,0], bool)).all())
        self.assertTrue(((cf.contains(999)==c).array ==
                         numpy.array([0,0,0,0,0,0,0,0,0], bool)).all())


    def test_Query(self):
        f = cf.read(self.filename)[0]

        q = cf.Query('wi', [2, 5])
        r = cf.Query('le', 67)
        s = q | r
        t = cf.Query('gt', 12, attr='bounds')
        u = s & t

        _ = repr(q)
        _ = repr(s)
        _ = repr(t)
        _ = repr(u)
        _ = str(q)
        _ = str(s)
        _ = str(t)
        _ = str(u)
        _ = u.dump(display=False)

        _ = u.attr
        _ = u.operator
        _ = q.attr
        _ = q.operator
        _ = q.value
        with self.assertRaises(Exception):
            _ = u.value

        self.assertTrue(u.equals(u.copy(), verbose=True))
        self.assertFalse(u.equals(t, verbose=False))

        _ = copy.deepcopy(u)

        c = f.dimension_coordinate('X')
        self.assertTrue(((cf.contains(21.1)==c).array ==
                         numpy.array([0,1,0,0,0,0,0,0,0], bool)).all())
        self.assertTrue(((cf.contains(999)==c).array ==
                         numpy.array([0,0,0,0,0,0,0,0,0], bool)).all())

        _ = cf.cellsize(34)
        _ = cf.cellsize(q)

        _ = cf.celllt(3)
        _ = cf.cellle(3)
        _ = cf.cellge(3)
        _ = cf.cellgt(3)
        _ = cf.cellwi(1, 2)
        _ = cf.cellwo(1, 2)

    def test_Query_object_units(self):
        """Check units are processed correctly in and from queries."""

        equivalent_units = {
            1000: ('m', 'km'),
            60: ('s', 'minute')
        }  # keys are conversion factors; only use exact equivalents for test
        for conversion, equivalents in equivalent_units.items():
            s_unit, l_unit = equivalents  # s for smaller, l for larger

            # Case 1: only specify units to one component
            q1 = cf.Query('gt', cf.Data(1, s_unit))
            q2 = cf.Query('ge', 1, units=s_unit)
            # Case 2: specify *equal* units across the two components
            q3 = cf.Query('lt', cf.Data(1, s_unit), units=s_unit)
            # Case 3: specify *equivalent* units across the two components
            q4 = cf.Query('le', cf.Data(1, s_unit), units=l_unit)
            # See also final Case 4, below.

            # A) test processed correctly inside unit itself
            for q in [q1, q2, q3, q4]:
                self.assertIsInstance(q, cf.query.Query)
                # TODO: should q4 should return s_ or l_unit? ATM is s_unit.
                self.assertTrue(q._value.Units._units == s_unit)

            with self.assertRaises(ValueError):
                # Case 4: provide non-equivalent units i.e. non-sensical query
                cf.Query('le', cf.Data(1, 'm'), units='s')

            # B) Check units are processed correctly when a Query is evaluated
            q5 = cf.Query('eq', conversion, units=s_unit)
            # Pre-comparison, values should be converted for consistent units
            self.assertTrue(
                q5.evaluate(cf.Data([1, 2], l_unit)).equals(
                    cf.Data([True, False]))
            )

    def test_Query_as_where_condition(self):
        """Check queries work correctly as conditions in 'where' method."""
        # TODO: extend test; added as-is to capture a specific bug (now fixed)

        s_data = cf.Data([30, 60, 90], 'second')

        m_lt_query = cf.lt(1, units="minute")
        s_lt_query = cf.lt(60, units="second")
        m_ge_query = cf.ge(1, units="minute")
        s_ge_query = cf.ge(60, units="second")
        for query_pair in [
                (m_lt_query, s_lt_query), (m_ge_query, s_ge_query)]:
            m_query, s_query = query_pair

            equal_units_where = s_data.data.where(s_query, 0)
            mixed_units_where = s_data.data.where(m_query, 0)
            self.assertTrue(
                (mixed_units_where.array == equal_units_where.array).all()
            )

            equal_units_where_masked = s_data.data.where(s_query, cf.masked)
            mixed_units_where_masked = s_data.data.where(m_query, cf.masked)
            self.assertEqual(
                mixed_units_where_masked.count(),
                equal_units_where_masked.count()
            )

    def test_Query_datetime1(self):
        d = cf.Data([[1., 5.], [6, 2]], 'days since 2000-12-29 21:00:00', calendar='standard')

        message = 'Diff ='+str( (d-cf.Data(cf.dt('2001-01-03 21:00:00', calendar='standard'))).array)

        self.assertTrue((d==cf.eq(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, True], [False, False]]), verbose=True), message)
        self.assertTrue((d==cf.ne(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, False], [True, True]])), message)
        self.assertTrue((d==cf.ge(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, True], [True, False]])), message)
        self.assertTrue((d==cf.gt(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, False], [True, False]])), message)
        self.assertTrue((d==cf.le(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, True], [False, True]])), message)
        self.assertTrue((d==cf.lt(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, False], [False, True]])), message)
        self.assertTrue((d==cf.wi(cf.dt('2000-12-31 21:00:00'), cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, True], [False, True]])), message)
        self.assertTrue((d==cf.wo(cf.dt('2000-12-31 21:00:00'), cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, False], [True, False]])), message)
        self.assertTrue((d==cf.set([cf.dt('2000-12-31 21:00:00'), cf.dt('2001-01-03 21:00:00')])).equals(cf.Data([[False, True], [False, True]])), message)

        _ = cf.seasons()
        [cf.seasons(n, start) for n in [1, 2, 3, 4, 6, 12] for start in range(1, 13)]
        with self.assertRaises(Exception):
            cf.seasons(13)
        with self.assertRaises(Exception):
            cf.seasons(start=8.456)

        _ = cf.mam()
        _ = cf.djf()
        _ = cf.jja()
        _ = cf.son()


    def test_Query_year_month_day_hour_minute_second(self):
        d = cf.Data([[1., 5.], [6, 2]], 'days since 2000-12-29 21:57:57', calendar='gregorian')

        self.assertTrue((d==cf.year(2000)).equals(cf.Data([[True, False], [False, True]])))
        self.assertTrue((d==cf.month(12)).equals(cf.Data([[True, False], [False, True]])))
        self.assertTrue((d==cf.day(3)).equals(cf.Data([[False, True], [False, False]])))
        d = cf.Data([[1., 5], [6, 2]], 'hours since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.hour(2)).equals(cf.Data([[False, True], [False, False]])))
        d = cf.Data([[1., 5], [6, 2]], 'minutes since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.minute(2)).equals(cf.Data([[False, True], [False, False]])))
        d = cf.Data([[1., 5], [6, 2]], 'seconds since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.second(2)).equals(cf.Data([[False, True], [False, False]])))

        d = cf.Data([[1., 5.], [6, 2]], 'days since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.year(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))
        self.assertTrue((d==cf.month(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))
        self.assertTrue((d==cf.day(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))
        d = cf.Data([[1., 5], [6, 2]], 'hours since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.hour(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))
        d = cf.Data([[1., 5], [6, 2]], 'minutes since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.minute(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))
        d = cf.Data([[1., 5], [6, 2]], 'seconds since 2000-12-29 21:57:57')
        self.assertTrue((d==cf.second(cf.ne(-1))).equals(cf.Data([[True, True], [True, True]])))


    def test_Query_dteq_dtne_dtge_dtgt_dtle_dtlt(self):
        d = cf.Data([[1., 5.], [6, 2]], 'days since 2000-12-29 21:00:00')

        message = 'Diff ='+str((d-cf.Data(cf.dt('2001-01-03 21:00:00', calendar='standard'))).array)

        self.assertTrue((d==cf.eq(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, True], [False, False]])), message)
        self.assertTrue((d==cf.ne(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, False], [True, True]])), message)
        self.assertTrue((d==cf.ge(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, True], [True, False]])), message)
        self.assertTrue((d==cf.gt(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[False, False], [True, False]])), message)
        self.assertTrue((d==cf.le(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, True], [False, True]])), message)
        self.assertTrue((d==cf.lt(cf.dt('2001-01-03 21:00:00'))).equals(cf.Data([[True, False], [False, True]])), message)

        self.assertTrue((d==cf.eq(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[False, True], [False, False]])), message)
        self.assertTrue((d==cf.ne(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[True, False], [True, True]])), message)
        self.assertTrue((d==cf.ge(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[False, True], [True, False]])), message)
        self.assertTrue((d==cf.gt(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[False, False], [True, False]])), message)
        self.assertTrue((d==cf.le(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[True, True], [False, True]])), message)
        self.assertTrue((d==cf.lt(cf.dt(2001, 1, 3, 21, 0, 0))).equals(cf.Data([[True, False], [False, True]])), message)

        d = cf.dt(2002, 6, 16)
        self.assertFalse(cf.eq(cf.dt(1990, 1, 1)) == d)
        self.assertTrue(cf.eq(cf.dt(2002, 6, 16)) == d)
        self.assertFalse(cf.eq(cf.dt('2100-1-1')) == d)
        self.assertFalse(cf.eq(cf.dt('2001-1-1')) & cf.eq(cf.dt(2010, 12, 31)) == d)

        d = cf.dt(2002, 6, 16)
        self.assertTrue(cf.ge(cf.dt(1990, 1, 1)) == d)
        self.assertTrue(cf.ge(cf.dt(2002, 6, 16)) == d)
        self.assertFalse(cf.ge(cf.dt('2100-1-1')) == d)
        self.assertFalse(cf.ge(cf.dt('2001-1-1')) & cf.ge(cf.dt(2010, 12, 31)) == d)

        d = cf.dt(2002, 6, 16)
        self.assertTrue(cf.gt(cf.dt(1990, 1, 1)) == d)
        self.assertFalse(cf.gt(cf.dt(2002, 6, 16)) == d)
        self.assertFalse(cf.gt(cf.dt('2100-1-1')) == d)
        self.assertTrue(cf.gt(cf.dt('2001-1-1')) & cf.le(cf.dt(2010, 12, 31)) == d)

        d = cf.dt(2002, 6, 16)
        self.assertTrue(cf.ne(cf.dt(1990, 1, 1)) == d)
        self.assertFalse(cf.ne(cf.dt(2002, 6, 16)) == d)
        self.assertTrue(cf.ne(cf.dt('2100-1-1')) == d)
        self.assertTrue(cf.ne(cf.dt('2001-1-1')) & cf.ne(cf.dt(2010, 12, 31)) == d)

        d = cf.dt(2002, 6, 16)
        self.assertFalse(cf.le(cf.dt(1990, 1, 1)) == d)
        self.assertTrue(cf.le(cf.dt(2002, 6, 16)) == d)
        self.assertTrue(cf.le(cf.dt('2100-1-1')) == d)
        self.assertFalse(cf.le(cf.dt('2001-1-1')) & cf.le(cf.dt(2010, 12, 31)) == d)

        d = cf.dt(2002, 6, 16)
        self.assertFalse(cf.lt(cf.dt(1990, 1, 1)) == d)
        self.assertFalse(cf.lt(cf.dt(2002, 6, 16)) == d)
        self.assertTrue(cf.lt(cf.dt('2100-1-1')) == d)
        self.assertFalse(cf.lt(cf.dt('2001-1-1')) & cf.lt(cf.dt(2010, 12, 31)) == d)


    def test_Query_evaluate(self):
        for x in (5, cf.Data(5, 'kg m-2'), cf.Data([5], 'kg m-2 s-1')):
            self.assertTrue(x == cf.eq(5))
            self.assertTrue(x == cf.lt(8))
            self.assertTrue(x == cf.le(8))
            self.assertTrue(x == cf.gt(3))
            self.assertTrue(x == cf.ge(3))
            self.assertTrue(x == cf.wi(3, 8))
            self.assertTrue(x == cf.wo(8, 11))
            self.assertTrue(x == cf.set([3, 5, 8]))

            self.assertTrue(cf.eq(5)          == x)
            self.assertTrue(cf.lt(8)          == x)
            self.assertTrue(cf.le(8)          == x)
            self.assertTrue(cf.gt(3)          == x)
            self.assertTrue(cf.ge(3)          == x)
            self.assertTrue(cf.wi(3, 8)       == x)
            self.assertTrue(cf.wo(8, 11)      == x)
            self.assertTrue(cf.set([3, 5, 8]) == x)

            self.assertFalse(x == cf.eq(8))
            self.assertFalse(x == cf.lt(3))
            self.assertFalse(x == cf.le(3))
            self.assertFalse(x == cf.gt(8))
            self.assertFalse(x == cf.ge(8))
            self.assertFalse(x == cf.wi(8, 11))
            self.assertFalse(x == cf.wo(3, 8))
            self.assertFalse(x == cf.set([3, 8, 11]))

            self.assertFalse(x == cf.eq(8)           == x)
            self.assertFalse(x == cf.lt(3)           == x)
            self.assertFalse(x == cf.le(3)           == x)
            self.assertFalse(x == cf.gt(8)           == x)
            self.assertFalse(x == cf.ge(8)           == x)
            self.assertFalse(x == cf.wi(8, 11)       == x)
            self.assertFalse(x == cf.wo(3, 8)        == x)
            self.assertFalse(x == cf.set([3, 8, 11]) == x)
        # --- End: for

        c = cf.wi(2, 4)
        d = cf.wi(6, 8)

        e = d | c

        self.assertTrue(c.evaluate(3))
        self.assertFalse(c.evaluate(5))

        self.assertTrue(e.evaluate(3))
        self.assertTrue(e.evaluate(7))
        self.assertFalse(e.evaluate(5))

        self.assertTrue(3 == c)
        self.assertFalse(5 == c)

        self.assertTrue(c == 3)
        self.assertFalse(c == 5)

        self.assertTrue(3 == e)
        self.assertTrue(7 == e)
        self.assertFalse(5 == e)

        self.assertTrue(e == 3)
        self.assertTrue(e == 7)
        self.assertFalse(e == 5)


        x = 'qwerty'
        self.assertTrue(x == cf.eq('qwerty'))
        self.assertTrue(x == cf.eq(re.compile('^qwerty$')))
        self.assertTrue(x == cf.eq(re.compile('qwe')))
        self.assertTrue(x == cf.eq(re.compile('qwe.*')))
        self.assertTrue(x == cf.eq(re.compile('.*qwe')))
        self.assertTrue(x == cf.eq(re.compile('.*rty')))
        self.assertTrue(x == cf.eq(re.compile('.*rty$')))
        self.assertTrue(x == cf.eq(re.compile('^.*rty$')))
        self.assertTrue(x == cf.eq(re.compile('rty$')))

        self.assertTrue(x != cf.eq('QWERTY'))
        self.assertTrue(x != cf.eq(re.compile('QWERTY'  )))
        self.assertTrue(x != cf.eq(re.compile('^QWERTY$')))
        self.assertTrue(x != cf.eq(re.compile('QWE'     )))
        self.assertTrue(x != cf.eq(re.compile('QWE.*'   )))
        self.assertTrue(x != cf.eq(re.compile('.*QWE'   )))
        self.assertTrue(x != cf.eq(re.compile('.*RTY'   )))
        self.assertTrue(x != cf.eq(re.compile('.*RTY$'  )))
        self.assertTrue(x != cf.eq(re.compile('^.*RTY$' )))


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import inspect
import os
import sys
import unittest

import cf

class CellMethodTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    strings = ('t: mean',
               'time: point',
               'time: maximum',
               'time: sum',
               'lon: maximum time: mean',
               'time: mean lon: maximum',
               'lat: lon: standard_deviation',
               'lon: standard_deviation lat: standard_deviation',
               'time: standard_deviation (interval: 1 day)',
               'area: mean',
               'lon: lat: mean',
               'time: variance (interval: 1 hr comment: sampled instantaneously)',
               'time: mean',
               'time: mean time: maximum',
               'time: mean within years time: maximum over years',
               'time: mean within days time: maximum within years time: variance over years',
               'time: standard_deviation (interval: 1 day)',
               'time: standard_deviation (interval: 1 year)',
               'time: standard_deviation (interval: 30 year)',
               'time: standard_deviation (interval: 1.0 year)',
               'time: standard_deviation (interval: 30.0 year)',
               'lat: lon: standard_deviation (interval: 10 km)',
               'lat: lon: standard_deviation (interval: 10 km interval: 10 km)',
               'lat: lon: standard_deviation (interval: 0.1 degree_N interval: 0.2 degree_E)',
               'lat: lon: standard_deviation (interval: 0.123 degree_N interval: 0.234 degree_E)',
               'time: variance (interval: 1 hr comment: sampled instantaneously)',
               'area: mean where land',
               'area: mean where land_sea',
               'area: mean where sea_ice over sea',
               'area: mean where sea_ice over sea',
               'time: minimum within years time: mean over years',
               'time: sum within years time: mean over years',
               'time: mean within days time: mean over days',
               'time: minimum within days time: sum over days',
               'time: minimum within days time: maximum over days',
               'time: mean within days',
               'time: sum within days time: maximum over days',
           )

    test_only = []
#    test_only = ['test_CellMethod___str__']
#    test_only = ['test_CellMethod_equals']
#    test_only = ['test_CellMethod_equivalent']
#    test_only = ['test_CellMethod_get_set_delete']

    def test_CellMethod__repr__str__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for s in self.strings:
            cms = cf.CellMethod.create(s)
            t = ' '.join(map(str, cms))
            self.assertTrue(t == s, '{!r} != {!r}'.format(t, s))
            for cm in cms:
                _ = repr(cm)


    def test_CellMethod_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for s in self.strings:
            cms = cf.CellMethod.create(s)
            for cm in cms:
                self.assertTrue(cm.equals(cm.copy(), verbose=True))


    def test_CellMethod_equivalent(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for s in self.strings:
            cms = cf.CellMethod.create(s)
            for cm in cms:
                self.assertTrue(cm.equivalent(cm.copy(), verbose=True))
        # --- End: for

        # Intervals
        for s0, s1 in (
                ['lat: lon: mean (interval: 10 km)',
                 'lat: lon: mean (interval: 10 km)'],
                ['lat: lon: mean (interval: 10 km)',
                 'lat: lon: mean (interval: 10 km interval: 10 km)'],
                ['lat: lon: mean (interval: 10 km interval: 10 km)',
                 'lat: lon: mean (interval: 10 km interval: 10 km)'],
                ['lat: lon: mean (interval: 20 km interval: 10 km)',
                 'lat: lon: mean (interval: 20 km interval: 10 km)'],
                ['lat: lon: mean (interval: 20 km interval: 10 km)',
                 'lat: lon: mean (interval: 20000 m interval: 10000 m)'],

                ['lat: lon: mean (interval: 10 km)',
                 'lon: lat: mean (interval: 10 km)'],
                ['lat: lon: mean (interval: 10 km)',
                 'lon: lat: mean (interval: 10 km interval: 10 km)'],
                ['lat: lon: mean (interval: 10 km interval: 10 km)',
                 'lon: lat: mean (interval: 10 km interval: 10 km)'],
                ['lat: lon: mean (interval: 20 km interval: 10 km)',
                 'lon: lat: mean (interval: 10 km interval: 20 km)'],
                ['lat: lon: mean (interval: 20 km interval: 10 km)',
                 'lon: lat: mean (interval: 10000 m interval: 20000 m)'],
        ):
            cms0 = cf.CellMethod.create(s0)
            cms1 = cf.CellMethod.create(s1)

            for cm0, cm1 in zip(cms0, cms1):
                self.assertTrue(cm0.equivalent(cm1, verbose=True),
                                '{0!r} not equivalent to {1!r}'.format(cm0, cm1))
        # --- End: for


    def test_CellMethod_get_set_delete(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        cm0, cm1 = cf.CellMethod.create('time: minimum within days time: sum over years')

        self.assertTrue(cm0.within == 'days')
        self.assertTrue(cm1.get_qualifier('within', None) is None)
        self.assertTrue(cm0.get_qualifier('where', None) is None)
        self.assertTrue(cm0.get_qualifier('over', None) is None)
        self.assertTrue(cm1.over == 'years')
        self.assertTrue(cm0.method == 'minimum')
        self.assertTrue(cm1.method == 'sum')
        self.assertTrue(cm0.axes == ('time',))
        self.assertTrue(cm1.axes == ('time',))


    def test_CellMethod_intervals(self):
        cm = cf.CellMethod.create('lat: mean (interval: 1 hour)')[0]

        self.assertEqual('1 hour', str(cm.intervals[0]))


# --- End: class


if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm

from . import mixin

from .decorators import _deprecated_kwarg_check


class CellMeasure(mixin.PropertiesData,
                  cfdm.CellMeasure):
    '''A cell measure construct of the CF data model.

    A cell measure construct provides information that is needed about
    the size or shape of the cells and that depends on a subset of the
    domain axis constructs. Cell measure constructs have to be used
    when the size or shape of the cells cannot be deduced from the
    dimension or auxiliary coordinate constructs without special
    knowledge that a generic application cannot be expected to have.

    The cell measure construct consists of a numeric array of the
    metric data which spans a subset of the domain axis constructs,
    and properties to describe the data. The cell measure construct
    specifies a "measure" to indicate which metric of the space it
    supplies, e.g. cell horizontal areas, and must have a units
    property consistent with the measure, e.g. square metres. It is
    assumed that the metric does not depend on axes of the domain
    which are not spanned by the array, along which the values are
    implicitly propagated. CF-netCDF cell measure variables correspond
    to cell measure constructs.

    **NetCDF interface**

    The netCDF variable name of the construct may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
    `nc_has_variable` methods.

    '''
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def ismeasure(self):
        '''Always True.

    .. seealso:: `isauxiliary`, `isdimension`

    **Examples:**

    >>> c.ismeasure
    True

        '''
        return True

    @property
    def measure(self):
        '''TODO
        '''
        return self.get_measure(default=AttributeError())

    @measure.setter
    def measure(self, value): self.set_measure(value)

    @measure.deleter
    def measure(self):        self.del_measure(default=AttributeError())

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def creation_commands(self, representative_data=False,
                          namespace='cf', indent=0, string=True,
                          name='c', data_name='d'):
        '''Return the commands that would create the cell measure construct.

    .. versionadded:: 3.2.0

    .. seealso:: `cf.Data.creation_commands`,
                 `cf.Field.creation_commands`

    :Parameters:

        representative_data: `bool`, optional
            Return one-line representations of `Data` instances, which
            are not executable code but prevent the data being
            converted in its entirety to a string representation.

        namespace: `str`, optional
            The namespace containing classes of the ``cf``
            package. This is prefixed to the class name in commands
            that instantiate instances of ``cf`` objects. By default,
            *namespace* is ``'cf'``, i.e. it is assumed that ``cf``
            was imported as ``import cf``.

            *Parameter example:*
              If ``cf`` was imported as ``import cf as cfp`` then set
              ``namespace='cfp'``

            *Parameter example:*
              If ``cf`` was imported as ``from cf import *`` then set
              ``namespace=''``

        indent: `int`, optional
            Indent each line by this many spaces. By default no
            indentation is applied. Ignored if *string* is False.

        string: `bool`, optional
            If False then return each command as an element of a
            `list`. By default the commands are concatenated into
            a string, with a new line inserted between each command.

    :Returns:

        `str` or `list`
            The commands in a string, with a new line inserted between
            each command. If *string* is False then the separate
            commands are returned as each element of a `list`.

    **Examples:**

        TODO

        '''
        out = super().creation_commands(
            representative_data=representative_data, indent=indent,
            namespace=namespace, string=False, name=name,
            data_name=data_name)

        measure = self.get_measure(None)
        if measure is not None:
            out.append("{}.set_measure({!r})".format(name, measure))

        if string:
            out[0] = indent+out[0]
            out = ('\n'+indent).join(out)

        return out

    @_deprecated_kwarg_check('relaxed_identity')
    def identity(self, default='', strict=None, relaxed=False,
                 nc_only=False, relaxed_identity=None):
        '''Return the canonical identity.

    By default the identity is the first found of the following:

    * The `measure` attribute, preceeded by ``'measure:'``.
    * The `standard_name` property.
    * The `id` attribute, preceeded by ``'id%'``.
    * The `long_name` property, preceeded by ``'long_name='``.
    * The netCDF variable name, preceeded by ``'ncvar%'``.
    * The value of the *default* parameter.

    .. versionadded:: 3.0.0

    .. seealso:: `id`, `identities`, `long_name`, `measure`,
                 `nc_get_variable`, `standard_name`

    :Parameters:

        default: optional
            If no identity can be found then return the value of the
            default parameter.

        strict: `bool`, optional
            If True then the identity is the first found of only the
            "measure" arttribute, "standard_name" property or the "id"
            attribute.

        relaxed: `bool`, optional
            If True then the identity is the first found of only the
            "measure" arttribute, the "standard_name" property, the
            "id" attribute, the "long_name" property or the netCDF
            variable name.

        nc_only: `bool`, optional
            If True then only take the identity from the netCDF
            variable name.

    :Returns:

            The identity.

    **Examples:**

    >>> c.measure
    'area'
    >>> c.properties()
    {'long_name': 'cell_area',
     'foo': 'bar'}
    >>> c.nc_get_variable()
    'areacello'
    >>> c.identity()
    'measure:area'
    >>> del c.measure
    >>> c.identity()
    'long_name=cell_area'
    >>> del c.long_name
    >>> c.identity()
    'ncvar%areacello'
    >>> c.nc_del_variable()
    'areacello'
    >>> c.identity()
    ''
    >>> c.identity('no identity')
    'no identity'

        '''
        err_msg = "%{0} and 'nc_only' parameters cannot both be True"
        if nc_only:
            if strict:
                raise ValueError(err_msg.format("'strict'"))

            if relaxed:
                raise ValueError(err_msg.format("'relaxed'"))

            n = self.nc_get_variable(None)
            if n is not None:
                return 'ncvar%{0}'.format(n)

            return default

        n = self.get_measure(default=None)
        if n is not None:
            return 'measure:{0}'.format(n)

        n = self.get_property('standard_name', None)
        if n is not None:
            return '{0}'.format(n)

        n = getattr(self, 'id', None)
        if n is not None:
            return 'id%{0}'.format(n)

        if relaxed:
            n = self.get_property('long_name', None)
            if n is not None:
                return 'long_name={0}'.format(n)

            n = self.nc_get_variable(None)
            if n is not None:
                return 'ncvar%{0}'.format(n)

            return default

        if strict:
            return default

        for prop in ('long_name',):
            n = self.get_property(prop, None)
            if n is not None:
                return '{0}={1}'.format(prop, n)
        # --- End: for

        n = self.nc_get_variable(None)
        if n is not None:
            return 'ncvar%{0}'.format(n)

        return default


# --- End: class



### DEPRECATED at 3.2.0 ###

from .. import mixin

from ..data.data import Data


class Coordinate(mixin.PropertiesDataBounds):
    '''Mixin class for dimension or auxiliary coordinate constructs.

    .. versionadded:: 3.0.0

    '''
    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def ctype(self):
        '''The CF coordinate type.

    One of ``'T'``, ``'X'``, ``'Y'`` or ``'Z'`` if the coordinate
    construct is for the respective CF axis type, otherwise `None`.

    .. seealso:: `T`, `X`, `~cf.Coordinate.Y`, `Z`

    **Examples:**

    >>> c.X
    True
    >>> c.ctype
    'X'

    >>> c.T
    True
    >>> c.ctype
    'T'

        '''
        for t in ('T', 'X', 'Y', 'Z'):
            if getattr(self, t):
                return t

    @property
    def T(self):
        '''True if and only if the data are coordinates for a CF 'T' axis.

    CF 'T' axis coordinates are defined by having one or more of the
    following:

      * The `axis` property has the value ``'T'``
      * Units of latitude

    .. seealso:: `ctype`, `X`, `~cf.Coordinate.Y`, `Z`

    **Examples:**

    >>> c.Units
    <CF Units: seconds since 1992-10-8>
    >>> c.T
    True

        '''
        return self.Units.isreftime or self.get_property('axis', None) == 'T'

    @property
    def X(self):
        '''True if and only if the data are coordinates for a CF 'X' axis.

    CF 'X' axis coordinates are defined by having one or more of the
    following:

      * The `axis` property has the value ``'X'``
      * Units of longitude
      * The `standard_name` property is one of ``'longitude'``,
        ``'projection_x_coordinate'`` or ``'grid_longitude'``

    .. seealso:: `ctype`, `T`, `~cf.Coordinate.Y`, `Z`

    **Examples:**

    >>> c.Units
    <CF Units: degreeE>
    >>> c.X
    True

    >>> c.standard_name
    'longitude'
    >>> c.X
    True

    >>> c.axis == 'X' and c.X
    True

        '''
#        data  = self.get_data(None)
#        if data is not None and data.ndim > 1:
#            return self.get_property('axis', None) == 'X'

        return (
            self.Units.islongitude or
            self.get_property('axis', None) == 'X' or
            self.get_property('standard_name', None) in
            ('longitude',
             'projection_x_coordinate',
             'grid_longitude')
        )

    @property
    def Y(self):
        '''True if and only if the data are coordinates for a CF 'Y' axis.

    CF 'Y' axis coordinates are defined by having one or more of the
    following:

      * The `axis` property has the value ``'Y'``
      * Units of latitude
      * The `standard_name` property is one of ``'latitude'``,
        ``'projection_y_coordinate'`` or ``'grid_latitude'``

    .. seealso:: `ctype`, `T`, `X`, `Z`

    **Examples:**

    >>> c.Units
    <CF Units: degree_north>
    >>> c.Y
    True

    >>> c.standard_name == 'latitude'
    >>> c.Y
    True

        '''
#        if self.ndim > 1:
#            return self.get_property('axis', None) == 'Y'

        return (
            self.Units.islatitude or
            self.get_property('axis', None) == 'Y' or
            self.get_property('standard_name', None) in
            ('latitude',
             'projection_y_coordinate',
             'grid_latitude')
        )

    @property
    def Z(self):
        '''True if and only if the data are coordinates for a CF 'Z' axis.

    CF 'Z' axis coordinates are defined by having one or more of the
    following:

      * The `axis` property has the value ``'Z'``
      * Units of pressure, level, layer or sigma_level
      * The `positive` property has the value ``'up'`` or ``'down'``
        (case insensitive)
      * The `standard_name` property is one of
        ``'atmosphere_ln_pressure_coordinate'``,
        ``'atmosphere_sigma_coordinate'``,
        ``'atmosphere_hybrid_sigma_pressure_coordinate'``,
        ``'atmosphere_hybrid_height_coordinate'``,
        ``'atmosphere_sleve_coordinate``', ``'ocean_sigma_coordinate'``,
        ``'ocean_s_coordinate'``, ``'ocean_s_coordinate_g1'``,
        ``'ocean_s_coordinate_g2'``, ``'ocean_sigma_z_coordinate'`` or
        ``'ocean_double_sigma_coordinate'``

    .. seealso:: `ctype`, `T`, `X`, `~cf.Coordinate.Y`

    **Examples:**

    >>> c.Units
    <CF Units: Pa>
    >>> c.Z
    True

    >>> c.Units.equivalent(cf.Units('K')) and c.positive == 'up'
    True
    >>> c.Z
    True

    >>> c.axis == 'Z' and c.Z
    True

    >>> c.Units
    <CF Units: sigma_level>
    >>> c.Z
    True

    >>> c.standard_name
    'ocean_sigma_coordinate'
    >>> c.Z
    True

        '''
#        if self.ndim > 1:
#            return self.get_property('axis', None) == 'Z'

        units = self.Units
        return (
            units.ispressure or
            str(self.get_property('positive', 'Z')).lower() in
            ('up', 'down') or
            self.get_property('axis', None) == 'Z' or
            (units and units.units in ('level', 'layer' 'sigma_level')) or
            self.get_property('standard_name', None) in
            ('atmosphere_ln_pressure_coordinate',
             'atmosphere_sigma_coordinate',
             'atmosphere_hybrid_sigma_pressure_coordinate',
             'atmosphere_hybrid_height_coordinate',
             'atmosphere_sleve_coordinate',
             'ocean_sigma_coordinate',
             'ocean_s_coordinate',
             'ocean_s_coordinate_g1',
             'ocean_s_coordinate_g2',
             'ocean_sigma_z_coordinate',
             'ocean_double_sigma_coordinate')
        )

    # ----------------------------------------------------------------
    # CF properties
    # ----------------------------------------------------------------
    @property
    def axis(self):
        '''The axis CF property.

    The `axis` property may be used to specify the type of
    coordinates. It may take one of the values `'X'`, `'Y'`, `'Z'` or
    `'T'` which stand for a longitude, latitude, vertical, or time
    axis respectively. A value of `'X'`, `'Y'` or `'Z'` may also also
    used to identify generic spatial coordinates (the values `'X'` and
    `'Y'` being used to identify horizontal coordinates).

    **Examples:**

    >>> c.axis = 'Y'
    >>> c.axis
    'Y'
    >>> del c.axis

    >>> c.set_property('axis', 'T')
    >>> c.get_property('axis')
    'T'
    >>> c.del_property('axis')

        '''
        return self.get_property('axis', default=AttributeError())

    @axis.setter
    def axis(self, value):
        self.set_property('axis', value)

    @axis.deleter
    def axis(self):
        self.del_property('axis')

    @property
    def positive(self):
        '''The positive CF property.

    The direction of positive (i.e., the direction in which the
    coordinate values are increasing), whether up or down, cannot in
    all cases be inferred from the `units`. The direction of positive
    is useful for applications displaying the data. The `positive`
    attribute may have the value ``'up'`` or ``'down'`` (case
    insensitive).

    For example, if ocean depth coordinates encode the depth of the
    surface as 0 and the depth of 1000 meters as 1000 then the
    `postive` property will have the value `'down'`.

    **Examples:**

    >>> c.positive = 'up'
    >>> c.positive
    'up'
    >>> del c.positive

    >>> c.set_property('positive', 'down')
    >>> c.get_property('positive')
    'down'
    >>> c.del_property('positive')

        '''
        return self.get_property('positive', default=AttributeError())

    @positive.setter
    def positive(self, value):
        self.set_property('positive', value)
        self._direction = None

    @positive.deleter
    def positive(self):
        self.del_property('positive')
        self._direction = None

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
#     def period(self, *value):
#         '''Set the period for cyclic coordinates.
#
# :Parameters:
#
#     value: data-like or `None`, optional
#         The period. The absolute value is used.
#
#         {+data-like-scalar}
#
# :Returns:
#
#     out: `cf.Data` or `None`
#         The period prior to the change, or the current period if no
#         *value* was specified. In either case, None is returned if the
#         period had not been set previously.
#
# **Examples:**
#
# >>> print(c.period())
# None
# >>> c.Units
# <CF Units: degrees_east>
# >>> print(c.period(360))
# None
# >>> c.period()
# <CF Data: 360.0 'degrees_east'>
# >>> import math
# >>> c.period(cf.Data(2*math.pi, 'radians'))
# <CF Data: 360.0 degrees_east>
# >>> c.period()
# <CF Data: 6.28318530718 radians>
# >>> c.period(None)
# <CF Data: 6.28318530718 radians>
# >>> print(c.period())
# None
# >>> print(c.period(-360))
# None
# >>> c.period()
# <CF Data: 360.0 degrees_east>
#
#         '''
#         old = self._period
#         if old is not None:
#             old = old.copy()
#
#         if not value:
#             return old
#
#         value = value[0]
#
#         if value is not None:
#             value = Data.asdata(value)
#             units = value.Units
#             if not units:
#                 value = value.override_units(self.Units)
#             elif units != self.Units:
#                 if units.equivalent(self.Units):
#                     value.Units = self.Units
#                 else:
#                     raise ValueError(
# "Period units {!r} are not equivalent to coordinate units {!r}".format(
#     units, self.Units))
#             # --- End: if
#
#             value = abs(value)
#             value.dtype = float
#
#             if self.isdimension:
#                 #  Faster than `range`
#                 array = self.array
#                 r =  abs(array[-1] - array[0])
#             else:
#                 r = self.data.range().datum(0)
#
#             if r >= value.datum(0):
#                 raise ValueError(
# "The coordinate range {!r} is not less than the period {!r}".format(
#     range, value))
#         # --- End: if
#
#         self._period = value
#
#         return old


# --- End: class



import datetime
import inspect
import os
import tempfile
import unittest

import numpy

import cf


#def _make_contiguous_file_with_bounds(filename):
#    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
#
#    n.Conventions = 'CF-1.7'
#    n.featureType = 'timeSeries'
#
#    station = n.createDimension('station', 4)
#    obs     = n.createDimension('obs'    , 24)
#    name_strlen = n.createDimension('name_strlen', 8)
#    bounds  = n.createDimension('bounds', 2)
#
#    lon = n.createVariable('lon', 'f8', ('station',))
#    lon.standard_name = "longitude"
#    lon.long_name = "station longitude"
#    lon.units = "degrees_east"
#    lon.bounds = "lon_bounds"
#    lon[...] = [-23, 0, 67, 178]
#
#    lon_bounds = n.createVariable('lon_bounds', 'f8', ('station', 'bounds'))
#    lon_bounds[...] = [[-24, -22],
#                       [ -1,   1],
#                       [ 66,  68],
#                       [177, 179]]
#
#    lat = n.createVariable('lat', 'f8', ('station',))
#    lat.standard_name = "latitude"
#    lat.long_name = "station latitude"
#    lat.units = "degrees_north"
#    lat[...] = [-9, 2, 34, 78]
#
#    alt = n.createVariable('alt', 'f8', ('station',))
#    alt.long_name = "vertical distance above the surface"
#    alt.standard_name = "height"
#    alt.units = "m"
#    alt.positive = "up"
#    alt.axis = "Z"
#    alt[...] = [0.5, 12.6, 23.7, 345]
#
#    station_name = n.createVariable('station_name', 'S1',
#                                    ('station', 'name_strlen'))
#    station_name.long_name = "station name"
#    station_name.cf_role = "timeseries_id"
#    station_name[...] = numpy.array([[x for x in 'station1'],
#                                     [x for x in 'station2'],
#                                     [x for x in 'station3'],
#                                     [x for x in 'station4']])
#
#    station_info = n.createVariable('station_info', 'i4', ('station',))
#    station_info.long_name = "some kind of station info"
#    station_info[...] = [-10, -9, -8, -7]
#
#    row_size = n.createVariable('row_size', 'i4', ('station',))
#    row_size.long_name = "number of observations for this station"
#    row_size.sample_dimension = "obs"
#    row_size[...] = [3, 7, 5, 9]
#
#    time = n.createVariable('time', 'f8', ('obs',))
#    time.standard_name = "time"
#    time.long_name = "time of measurement"
#    time.units = "days since 1970-01-01 00:00:00"
#    time.bounds = "time_bounds"
#    time[ 0: 3] = [-3, -2, -1]
#    time[ 3:10] = [1, 2, 3, 4, 5, 6, 7]
#    time[10:15] = [0.5, 1.5, 2.5, 3.5, 4.5]
#    time[15:24] = range(-2, 7)
#
#    time_bounds = n.createVariable('time_bounds', 'f8', ('obs', 'bounds'))
#    time_bounds[..., 0] = time[...] - 0.5
#    time_bounds[..., 1] = time[...] + 0.5
#
#    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
#    humidity.standard_name = "specific_humidity"
#    humidity.coordinates = "time lat lon alt station_name station_info"
#    humidity[ 0: 3] = numpy.arange(0, 3)
#    humidity[ 3:10] = numpy.arange(1, 71, 10)
#    humidity[10:15] = numpy.arange(2, 502, 100)
#    humidity[15:24] = numpy.arange(3, 9003, 1000)
#
#    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
#    temp.standard_name = "air_temperature"
#    temp.units = "Celsius"
#    temp.coordinates = "time lat lon alt station_name station_info"
#    temp[...] = humidity[...] + 273.15
#
#    n.close()
#
#    return filename
#
#
#def _make_contiguous_file(filename):
#    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
#
#    n.Conventions = 'CF-1.7'
#    n.featureType = 'timeSeries'
#
#    station = n.createDimension('station', 4)
#    obs     = n.createDimension('obs'    , 24)
#    name_strlen = n.createDimension('name_strlen', 8)
#
#    lon = n.createVariable('lon', 'f8', ('station',))
#    lon.standard_name = "longitude"
#    lon.long_name = "station longitude"
#    lon.units = "degrees_east"
#    lon[...] = [-23, 0, 67, 178]
#
#    lat = n.createVariable('lat', 'f8', ('station',))
#    lat.standard_name = "latitude"
#    lat.long_name = "station latitude"
#    lat.units = "degrees_north"
#    lat[...] = [-9, 2, 34, 78]
#
#    alt = n.createVariable('alt', 'f8', ('station',))
#    alt.long_name = "vertical distance above the surface"
#    alt.standard_name = "height"
#    alt.units = "m"
#    alt.positive = "up"
#    alt.axis = "Z"
#    alt[...] = [0.5, 12.6, 23.7, 345]
#
#    station_name = n.createVariable('station_name', 'S1',
#                                    ('station', 'name_strlen'))
#    station_name.long_name = "station name"
#    station_name.cf_role = "timeseries_id"
#    station_name[...] = numpy.array([[x for x in 'station1'],
#                                     [x for x in 'station2'],
#                                     [x for x in 'station3'],
#                                     [x for x in 'station4']])
#
#    station_info = n.createVariable('station_info', 'i4', ('station',))
#    station_info.long_name = "some kind of station info"
#    station_info[...] = [-10, -9, -8, -7]
#
#    row_size = n.createVariable('row_size', 'i4', ('station',))
#    row_size.long_name = "number of observations for this station"
#    row_size.sample_dimension = "obs"
#    row_size[...] = [3, 7, 5, 9]
#
#    time = n.createVariable('time', 'f8', ('obs',))
#    time.standard_name = "time"
#    time.long_name = "time of measurement"
#    time.units = "days since 1970-01-01 00:00:00"
#    time[ 0: 3] = [-3, -2, -1]
#    time[ 3:10] = [1, 2, 3, 4, 5, 6, 7]
#    time[10:15] = [0.5, 1.5, 2.5, 3.5, 4.5]
#    time[15:24] = range(-2, 7)
#
#    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
#    humidity.standard_name = "specific_humidity"
#    humidity.coordinates = "time lat lon alt station_name station_info"
#    humidity[ 0: 3] = numpy.arange(0, 3)
#    humidity[ 3:10] = numpy.arange(1, 71, 10)
#    humidity[10:15] = numpy.arange(2, 502, 100)
#    humidity[15:24] = numpy.arange(3, 9003, 1000)
#
#    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
#    temp.standard_name = "air_temperature"
#    temp.units = "Celsius"
#    temp.coordinates = "time lat lon alt station_name station_info"
#    temp[...] = humidity[...] + 273.15
#
#    n.close()
#
#    return filename
#
#
#def _make_indexed_file(filename):
#    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
#
#    n.Conventions = 'CF-1.7'
#    n.featureType = 'timeSeries'
#
#    station = n.createDimension('station', 4)
#    obs     = n.createDimension('obs'    , None)
#    name_strlen = n.createDimension('name_strlen', 8)
#
#    lon = n.createVariable('lon', 'f8', ('station',))
#    lon.standard_name = "longitude"
#    lon.long_name = "station longitude"
#    lon.units = "degrees_east"
#    lon[...] = [-23, 0, 67, 178]
#
#    lat = n.createVariable('lat', 'f8', ('station',))
#    lat.standard_name = "latitude"
#    lat.long_name = "station latitude"
#    lat.units = "degrees_north"
#    lat[...] = [-9, 2, 34, 78]
#
#    alt = n.createVariable('alt', 'f8', ('station',))
#    alt.long_name = "vertical distance above the surface"
#    alt.standard_name = "height"
#    alt.units = "m"
#    alt.positive = "up"
#    alt.axis = "Z"
#    alt[...] = [0.5, 12.6, 23.7, 345]
#
#    station_name = n.createVariable('station_name', 'S1',
#                                    ('station', 'name_strlen'))
#    station_name.long_name = "station name"
#    station_name.cf_role = "timeseries_id"
#    station_name[...] = numpy.array([[x for x in 'station1'],
#                                     [x for x in 'station2'],
#                                     [x for x in 'station3'],
#                                     [x for x in 'station4']])
#
#    station_info = n.createVariable('station_info', 'i4', ('station',))
#    station_info.long_name = "some kind of station info"
#    station_info[...] = [-10, -9, -8, -7]
#
#    #row_size[...] = [3, 7, 5, 9]
#    stationIndex = n.createVariable('stationIndex', 'i4', ('obs',))
#    stationIndex.long_name = "which station this obs is for"
#    stationIndex.instance_dimension= "station"
#    stationIndex[...] = [3, 2, 1, 0, 2, 3, 3, 3, 1, 1, 0, 2,
#                         3, 1, 0, 1, 2, 3, 2, 3, 3, 3, 1, 1]
#
#    t = [[-3, -2, -1],
#         [1, 2, 3, 4, 5, 6, 7],
#         [0.5, 1.5, 2.5, 3.5, 4.5],
#         range(-2, 7)]
#
#    time = n.createVariable('time', 'f8', ('obs',))
#    time.standard_name = "time"
#    time.long_name = "time of measurement"
#    time.units = "days since 1970-01-01 00:00:00"
#    ssi = [0, 0, 0, 0]
#    for i, si in enumerate(stationIndex[...]):
#        time[i] = t[si][ssi[si]]
#        ssi[si] += 1
#
#    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
#    humidity.standard_name = "specific_humidity"
#    humidity.coordinates = "time lat lon alt station_name station_info"
#
#    h = [numpy.arange(0, 3),
#         numpy.arange(1, 71, 10),
#         numpy.arange(2, 502, 100),
#         numpy.arange(3, 9003, 1000)]
#
#    ssi = [0, 0, 0, 0]
#    for i, si in enumerate(stationIndex[...]):
#        humidity[i] = h[si][ssi[si]]
#        ssi[si] += 1
#
#    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
#    temp.standard_name = "air_temperature"
#    temp.units = "Celsius"
#    temp.coordinates = "time lat lon alt station_name station_info"
#    temp[...] = humidity[...] + 273.15
#
#    n.close()
#
#    return filename
#
#
#def _make_indexed_contiguous_file(filename):
#    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
#
#    n.Conventions = 'CF-1.6'
#    n.featureType = "timeSeriesProfile"
#
#    # 3 stations
#    station = n.createDimension('station', 3)
#    # 58 profiles spreadover 4 stations, each at a different time
#    profile = n.createDimension('profile', 58)
#    obs      = n.createDimension('obs'    , None) #
#    name_strlen = n.createDimension('name_strlen', 8)
#
#    lon = n.createVariable('lon', 'f8', ('station',))
#    lon.standard_name = "longitude"
#    lon.long_name = "station longitude"
#    lon.units = "degrees_east"
#    lon[...] = [-23, 0, 67]
#
#    lat = n.createVariable('lat', 'f8', ('station',))
#    lat.standard_name = "latitude"
#    lat.long_name = "station latitude"
#    lat.units = "degrees_north"
#    lat[...] = [-9, 2, 34]
#
#    alt = n.createVariable('alt', 'f8', ('station',))
#    alt.long_name = "vertical distance above the surface"
#    alt.standard_name = "height"
#    alt.units = "m"
#    alt.positive = "up"
#    alt.axis = "Z"
#    alt[...] = [0.5, 12.6, 23.7]
#
#    station_name = n.createVariable('station_name', 'S1',
#                                    ('station', 'name_strlen'))
#    station_name.long_name = "station name"
#    station_name.cf_role = "timeseries_id"
#    station_name[...] = numpy.array([[x for x in 'station1'],
#                                     [x for x in 'station2'],
#                                     [x for x in 'station3']])
#
#    profile = n.createVariable('profile', 'i4', ('profile'))
#    profile.cf_role = "profile_id"
#    profile[...] = numpy.arange(58) + 100
#
#    station_info = n.createVariable('station_info', 'i4', ('station',))
#    station_info.long_name = "some kind of station info"
#    station_info[...] = [-10, -9, -8]
#
#    stationIndex = n.createVariable('stationIndex', 'i4', ('profile',))
#    stationIndex.long_name = "which station this profile is for"
#    stationIndex.instance_dimension= "station"
#    stationIndex[...] = [2, 1, 0, 2, 1, 1, 0, 2,
#                         1, 0, 1, 2, 2, 1, 1,
#                         2, 1, 0, 2, 1, 1, 0, 2,
#                         1, 0, 1, 2, 2, 1, 1,
#                         2, 1, 0, 2, 1, 1, 0, 2,
#                         1, 0, 1, 2, 2, 1, 1,
#                         2, 1, 0, 2, 1, 1, 0, 2,
#                         1, 0, 1, 2, 2]
#    # station N has list(stationIndex[...]).count(N) profiles
#
#    row_size = n.createVariable('row_size', 'i4', ('profile',))
#    row_size.long_name = "number of observations for this profile"
#    row_size.sample_dimension = "obs"
#    row_size[...] = [1, 4, 1, 3, 2, 2, 3, 3, 1, 2, 2, 3, 2, 2, 2, 2, 1, 2, 1, 3, 3, 2, 1,
#                     3, 1, 3, 2, 3, 1, 3, 3, 2, 2, 2, 1, 1, 1, 3, 1, 1, 2, 1, 1, 3, 3, 2,
#                     2, 2, 2, 1, 2, 3, 3, 3, 2, 3, 1, 1] # sum = 118
#
#    time = n.createVariable('time', 'f8', ('profile',))
#    time.standard_name = "time"
#    time.long_name = "time"
#    time.units = "days since 1970-01-01 00:00:00"
#    t0 = [3, 0, -3]
#    ssi = [0, 0, 0]
#    for i, si in enumerate(stationIndex[...]):
#        time[i] = t0[si] + ssi[si]
#        ssi[si] += 1
#
#    z = n.createVariable('z', 'f8', ('obs',))
#    z.standard_name = "altitude"
#    z.long_name = "height above mean sea level"
#    z.units = "km"
#    z.axis = "Z"
#    z.positive = "up"
#
##        z0 = [1, 0, 3]
##        i = 0
##        for s, r in zip(stationIndex[...], row_size[...]):
##            z[i:i+r] = z0[s] + numpy.sort(numpy.random.uniform(0, numpy.random.uniform(1, 2), r))
##            i += r
#
#    data =  [3.51977705293769, 0.521185292100177, 0.575154265863394,
#             1.08495843717095, 1.37710968624395, 2.07123455611723, 3.47064474274781,
#             3.88569849023813, 4.81069254279537, 0.264339600625496, 0.915704970094182,
#             0.0701532210336895, 0.395517651420933, 1.00657582854276,
#             1.17721374303641, 1.82189345615046, 3.52424307197668, 3.93200473199559,
#             3.95715099603671, 1.57047493027102, 1.09938982652955, 1.17768722826975,
#             0.251803399458277, 1.59673486865804, 4.02868944763605, 4.03749228832264,
#             4.79858281590985, 3.00019933315412, 3.65124061660449, 0.458463542157766,
#             0.978678197083262, 0.0561560792556281, 0.31182013232255,
#             3.33350065357286, 4.33143904011861, 0.377894196412131, 1.63020681064712,
#             2.00097025264771, 3.76948048424458, 0.572927165845568, 1.29408313557905,
#             1.81296270533192, 0.387142669131077, 0.693459187515738, 1.69261930636298,
#             1.38258797228361, 1.82590759889566, 3.34993297710761, 0.725250730922501,
#             1.38221693486728, 1.59828555215646, 1.59281225554253, 0.452340646918555,
#             0.976663373825433, 1.12640496317618, 3.19366847375422, 3.37209133117904,
#             3.40665008236976, 3.53525896684001, 4.10444186715724, 0.14920937817654,
#             0.0907197953552753, 0.42527916794473, 0.618685137936187,
#             3.01900591447357, 3.37205542289986, 3.86957342976163, 0.17175098751914,
#             0.990040375014957, 1.57011428605984, 2.12140567043994, 3.24374743730506,
#             4.24042441581785, 0.929509749153725, 0.0711997786817564,
#             2.25090028461898, 3.31520955860746, 3.49482624434274, 3.96812568493549,
#             1.5681807261767, 1.79993011515465, 0.068325990211909, 0.124469638352167,
#             3.31990436971169, 3.84766748039389, 0.451973490541035, 1.24303219956085,
#             1.30478004656262, 0.351892459787624, 0.683685812990457,
#             0.788883736575568, 3.73033428872491, 3.99479807507392, 0.811582011950481,
#             1.2241242448019, 1.25563109687369, 2.16603674712822, 3.00010622131408,
#             3.90637137662453, 0.589586644805982, 0.104656387266266,
#             0.961185900148304, 1.05120351477824, 1.29460917520233, 2.10139985693684,
#             3.64252693587415, 3.91197236350995, 4.56466622863717, 0.556476687600461,
#             0.783717448678148, 0.910917550635007, 1.59750076220451, 1.97101264162631,
#             0.714693043642084, 0.904381625638779, 1.03767817888021, 4.10124675852254,
#             3.1059214185543]
#    data = numpy.around(data, 2)
#    z[...] = data
#
#    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
#    humidity.standard_name = "specific_humidity"
#    humidity.coordinates = "time lat lon alt z station_name station_info profile"
#
#    data *= 10
#    data = numpy.around(data, 2)
#    humidity[...] = data
#
#    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
#    temp.standard_name = "air_temperature"
#    temp.units = "Celsius"
#    temp.coordinates = "time lat lon alt z station_name station_info profile"
#
#    data += 2731.5
#    data = numpy.around(data, 2)
#    temp[...] = data
#
#    n.close()
#
#    return filename
#
#
#contiguous_file = _make_contiguous_file_with_bounds('DSG_timeSeries_contiguous.nc')
#indexed_file    = _make_indexed_file('DSG_timeSeries_indexed.nc')
#indexed_contiguous_file = _make_indexed_contiguous_file('DSG_timeSeriesProfile_indexed_contiguous.nc')


class DSGTest(unittest.TestCase):
    def setUp(self):
        self.contiguous         = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                               'DSG_timeSeries_contiguous.nc')
        self.indexed            = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                               'DSG_timeSeries_indexed.nc')
        self.indexed_contiguous = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                               'DSG_timeSeriesProfile_indexed_contiguous.nc')

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cf_', dir='.')
        os.close(fd)

        a = numpy.ma.masked_all((4, 9), dtype=float)
        a[0, 0:3] = [0.0, 1.0, 2.0]
        a[1, 0:7] = [1.0, 11.0, 21.0, 31.0, 41.0, 51.0, 61.0]
        a[2, 0:5] = [2.0, 102.0, 202.0, 302.0, 402.0]
        a[3, 0:9] = [3.0, 1003.0, 2003.0, 3003.0, 4003.0, 5003.0, 6003.0, 7003.0, 8003.0]
        self.a = a

        b = numpy.array([[[20.7, -99, -99, -99],
                          [10.1, 11.8, 18.2, -99],
                          [11.0, 11.8, -99, -99],
                          [16.3, 20.0, -99, -99],
                          [13.8, 18.3, -99, -99],
                          [15.9 , -99, -99, -99],
                          [15.7, 21.2, -99, -99],
                          [22.5, -99, -99, -99],
                          [18.0, -99, -99, -99],
                          [12.6, 21.7, -99, -99],
                          [10.5, 12.9, 21.0, -99],
                          [16.0, 19.7, -99, -99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99]],

                         [[5.2, 5.8, 10.8, 13.8],
                          [2.6, 9.2, -99, -99],
                          [0.7, 4.0, -99, -99],
                          [15.7, -99, -99, -99],
                          [2.5, 16.0, -99, -99],
                          [4.6, 9.8, -99, -99],
                          [0.6, 3.1, -99, -99],
                          [3.8, -99, -99, -99],
                          [5.7, 12.9, 18.1, -99],
                          [3.9, 6.9, 16.9, -99],
                          [7.3, 13.8, 16.0, -99],
                          [4.5, 9.8, 11.3, -99],
                          [1.5, -99, -99, -99],
                          [0.9, 4.3, 6.2, -99],
                          [1.7, 9.9, -99, -99],
                          [9.3, -99, -99, -99],
                          [0.7, -99, -99, -99],
                          [15.7, -99, -99, -99],
                          [0.7, 1.2, -99, -99],
                          [4.5, 12.4, 13.0, -99],
                          [3.5, 6.8, 7.9, -99],
                          [8.1, 12.2, -99, -99],
                          [5.9, -99, -99, -99],
                          [1.0, 9.6, -99, -99],
                          [5.6, 7.8, 9.1, -99],
                          [7.1, 9.0, 10.4, -99]],

                         [[35.2, -99, -99, -99],
                          [34.7, 38.9, 48.1, -99],
                          [35.2, 39.3, 39.6, -99],
                          [40.3, 40.4, 48.0, -99],
                          [30.0, 36.5, -99, -99],
                          [33.3, 43.3, -99, -99],
                          [37.7, -99, -99, -99],
                          [33.5, -99, -99, -99],
                          [31.9, 33.7, -99, -99],
                          [34.1, 35.4, 41.0, -99],
                          [30.2, 33.7, 38.7, -99],
                          [32.4, 42.4, -99, -99],
                          [33.2, 34.9, 39.7, -99],
                          [33.2, -99, -99, -99],
                          [38.5, -99, -99, -99],
                          [37.3, 39.9, -99, -99],
                          [30.0, 39.1, -99, -99],
                          [36.4, 39.1, 45.6, -99],
                          [41.0, -99, -99, -99],
                          [31.1, -99, -99, -99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99]]])

        b = numpy.ma.where(b==-99, numpy.ma.masked, b)
        self.b = b

        self.test_only = []
#        self.test_only = ['test_DSG_indexed']


    def tearDown(self):
        os.remove(self.tempfilename)


    def test_DSG_contiguous(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.contiguous, verbose=False)

        self.assertTrue(len(f) == 2)

#        print ('\nf\n')
#        for x in f:
#            x.dump()

        # Select the specific humidity field
        q = [g for g in f
             if g.get_property('standard_name') == 'specific_humidity'][0]
#        f[0].data.inspect()
#        f[1].data.inspect()

        self.assertTrue(q._equals(q.data.array.mask, self.a.mask))

        self.assertTrue(q._equals(self.a, q.data.array),
                        '\nself.a=\n'+str(self.a)+'\nq.array=\n'+str(q.array))

        cf.write(f, self.tempfilename, verbose=False)
        g = cf.read(self.tempfilename)

#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=True))

        # ------------------------------------------------------------
        # Test creation
        # ------------------------------------------------------------
        # Define the ragged array values
        ragged_array = numpy.array([280, 282.5, 281, 279, 278, 279.5],
                                   dtype='float32')

        # Define the count array values
        count_array = [2, 4]

        # Create the count variable
        count_variable = cf.Count(data=cf.Data(count_array))
        count_variable.set_property('long_name', 'number of obs for this timeseries')

        # Create the contiguous ragged array object
        array = cf.RaggedContiguousArray(
            compressed_array=cf.Data(ragged_array),
            shape=(2, 4), size=8, ndim=2,
            count_variable=count_variable)

        # Create the field construct with the domain axes and the ragged
        # array
        tas = cf.Field()
        tas.set_properties({'standard_name': 'air_temperature',
                            'units': 'K',
                            'featureType': 'timeSeries'})

        # Create the domain axis constructs for the uncompressed array
        X = tas.set_construct(cf.DomainAxis(4))
        Y = tas.set_construct(cf.DomainAxis(2))

        # Set the data for the field
        tas.set_data(cf.Data(array), axes=[Y, X])

        cf.write(tas, self.tempfilename)


    def test_DSG_indexed(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.indexed)

        self.assertTrue(len(f) == 2)

        # Select the specific humidity field
        q = [g for g in f
             if g.get_property('standard_name') == 'specific_humidity'][0]

        self.assertTrue(q._equals(q.data.array.mask, self.a.mask))

        self.assertTrue(q._equals(q.data.array, self.a),
                        '\nself.a=\n'+str(self.a)+'\nq.array=\n'+str(q.array))


#        print ('\nf\n')
#        for x in f:
#            print(x)

        cf.write(f, self.tempfilename, verbose=False)
        g = cf.read(self.tempfilename)

#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=True))


    def test_DSG_indexed_contiguous(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.indexed_contiguous, verbose=False)

        self.assertTrue(len(f) == 2)

        # Select the specific humidity field
        q = f.select('specific_humidity')[0]
#        q = [g for g in f
#             if g.get_property('standard_name') == 'specific_humidity'][0]

        qa = q.data.array

#        print (qa[0, 12])
#        print (self.b[0, 12])

        for n in range(qa.shape[0]):
            for m in range(qa.shape[1]):
                self.assertTrue(q._equals(qa.mask[n, m], self.b.mask[n, m]),
                                str(n)+ ' '+ str(m)+ ' '+str(qa[n, m]) +' '+ str(self.b[n, m]))

        message= repr(qa-self.b) #+'\n'+repr(qa[2,0])+'\n'+repr(self.b[2, 0])
        self.assertTrue(q._equals(qa, self.b), message)

#        print ('\nf\n')
#        for x in f:
#            print(x)

        cf.write(f, self.tempfilename, verbose=False)
        g = cf.read(self.tempfilename, verbose=False)

#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=True))


    def test_DSG_create_contiguous(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Define the ragged array values
        ragged_array = numpy.array([1, 3, 4, 3, 6], dtype='float32')
        # Define the count array values
        count_array = [2, 3]

        # Initialise the count variable
        count_variable = cf.Count(data=cf.Data(count_array))
        count_variable.set_property('long_name', 'number of obs for this timeseries')

        # Initialise the contiguous ragged array object
        array = cf.RaggedContiguousArray(
            compressed_array=cf.Data(ragged_array),
            shape=(2, 3), size=6, ndim=2,
            count_variable=count_variable)

        # Initialize the auxiliary coordinate construct with the ragged
        # array and set some properties
        z = cf.AuxiliaryCoordinate(
            data=cf.Data(array),
            properties={'standard_name': 'height',
                        'units': 'km',
                        'positive': 'up'})

        self.assertTrue((z.data.array == numpy.ma.masked_array(
            data=[[1.0, 3.0, 99],
                  [4.0, 3.0, 6.0]],
            mask=[[False, False,  True],
                  [False, False, False]],
            fill_value=1e+20,
            dtype='float32')).all())

        self.assertTrue(z.data.get_compression_type() == 'ragged contiguous')

        self.assertTrue((z.data.compressed_array == numpy.array(
            [1., 3., 4., 3., 6.], dtype='float32')).all())

        self.assertTrue((z.data.get_count().data.array == numpy.array(
            [2, 3])).all())


# --- End: class


if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    print(cf.environment(display=False))
    print()
    unittest.main(verbosity=2)
